#![allow(dead_code)]

use std::iter::Peekable;

#[derive(PartialEq, Eq, Debug)]
pub struct TokenError {
    error: String,
    line: usize,
    col: usize,
}

#[derive(PartialEq, Eq, Debug)]
pub enum Token {
    Name(String),
    Id(String),
    Class(String),
    SlWhitespace(String),
    MlWhitespace(String),
    SlComment(String),
    MlComment(String),
    Int(String),
    Float(String),
    String(String),
    Null(String),
    True(String),
    False(String),
    LBrace(String),
    RBrace(String),
    LBracket(String),
    RBracket(String),
    Equal(String),
    Colon(String),
    SemiColon(String),
    Comma(String),
    Error(TokenError),
}

fn make_keyword_or_name<T: Iterator<Item = char>>(
    iter: &mut Peekable<T>,
    line: &mut usize,
) -> Token {
    let mut result: Vec<char> = Vec::new();
    loop {
        match iter.peek() {
            Some('A'..='Z' | 'a'..='z') => {
                *line += 1;
                let c = iter
                    .next()
                    .expect("iterator should not be mutated between peek & next");
                result.push(c);
            }
            Some('0'..='9') => {
                // we assert instead of returning an error because this means the lexer is written incorrectly
                assert_ne!(result.len(), 0);
                *line += 1;
                let c = iter
                    .next()
                    .expect("iterator should not be mutated between peek & next");
                result.push(c);
            }
            _ => {
                break match String::from_iter(result).as_str() {
                    s @ "null" => Token::Null(s.to_string()),
                    s @ "true" => Token::True(s.to_string()),
                    s @ "false" => Token::False(s.to_string()),
                    name => Token::Name(name.to_string()),
                }
            }
        }
    }
}

fn make_number<T: Iterator<Item = char>>(
    iter: &mut Peekable<T>,
    line: &mut usize,
    col: &mut usize,
) -> Result<Token, TokenError> {
    let mut result: Vec<char> = Vec::new();
    loop {
        let next = iter.peek();
        match next {
            Some('0'..='9') => {
                *line += 1;
                let c = iter
                    .next()
                    .expect("iterator should not be mutated between peek & next");
                result.push(c);
            }
            Some('.') => {
                *line += 1;
                if result.contains(&'.') {
                    iter.next();
                    return Err(TokenError {
                        error: "unexpected token".to_string(),
                        col: *col,
                        line: *line,
                    });
                } else {
                    iter.next()
                        .expect("iterator should not be mutated between peek & next");
                    result.push('.');
                }
            }
            _ => {
                break Ok(if result.contains(&'.') {
                    Token::Float(String::from_iter(result))
                } else {
                    Token::Int(String::from_iter(result))
                })
            }
        }
    }
}

fn make_string<T: Iterator<Item = char>>(
    iter: &mut T,
    line: &mut usize,
    col: &mut usize,
) -> Result<Token, TokenError> {
    let mut result: Vec<char> = Vec::new();
    let mut escaped = false;
    iter.next().expect("opening quote should exist");
    loop {
        let next = iter.next().ok_or(TokenError {
            error: "unexpected end of string".to_string(),
            line: *line,
            col: *col,
        })?;
        match next {
            '\\' => {
                *line += 1;
                escaped = !escaped;
                result.push('\\');
            }
            '"' => {
                *line += 1;
                if escaped {
                    result.push('"');
                    escaped = false;
                } else {
                    break Ok(Token::String(String::from_iter(result)));
                }
            }
            '\n' => {
                *line = 0;
                *col += 1;
                result.push('\n');
            }
            c => {
                *line += 1;
                escaped = false;
                result.push(c);
            }
        }
    }
}

fn lexer(code: &str) -> Vec<Token> {
    let mut tokens = Vec::new();
    let mut iter = code.chars().peekable();
    let mut col = 0;
    let mut line = 0;
    loop {
        let char = if let Some(c) = iter.peek() {
            c
        } else {
            break tokens;
        };

        match char {
            '"' => {
                let token = match make_string(&mut iter, &mut line, &mut col) {
                    Ok(token) => token,
                    Err(err) => Token::Error(err),
                };
                tokens.push(token);
            }
            '0'..='9' => {
                let token = match make_number(&mut iter, &mut line, &mut col) {
                    Ok(token) => token,
                    Err(err) => Token::Error(err),
                };
                tokens.push(token);
            }
            'a'..='z' | 'A'..='Z' => {
                let token = make_keyword_or_name(&mut iter, &mut line);
                tokens.push(token);
            }
            ' ' => {
                let mut result: Vec<char> = Vec::new();
                let token = loop {
                    let next = iter.peek();
                    match next {
                        Some(' ' | '\n' | '\r') => {
                            let c = iter
                                .next()
                                .expect("should not mutate between next & unwrap");
                            result.push(c);
                        }
                        _ => {
                            break if result.contains(&'\n') {
                                Token::MlWhitespace(String::from_iter(result))
                            } else {
                                Token::SlWhitespace(String::from_iter(result))
                            }
                        }
                    };
                };
                tokens.push(token);
            }
            _ => todo!(),
        }
    }
}

#[cfg(test)]
mod tests {
    use crate::bong::lexer::{lexer, Token};

    #[test]
    fn name_keyword_number_whitespace() {
        let result = lexer("abcd \n 1234   12.34 true false \"string\"");

        let space = |c: &str| -> Token { Token::SlWhitespace(c.to_string()) };

        assert_eq!(
            result,
            vec![
                Token::Name("abcd".to_string()),
                Token::MlWhitespace(" \n ".to_string()),
                Token::Int("1234".to_string()),
                space("   "),
                Token::Float("12.34".to_string()),
                space(" "),
                Token::True("true".to_string()),
                space(" "),
                Token::False("false".to_string()),
                space(" "),
                Token::String("string".to_string()),
            ]
        )
    }

    #[test]
    fn test_example_1() {
        let text = "text.title {
    // text { \"hello world\" }
    \"hello world\"
}";
        let tokens = lexer(text);
        assert_eq!(
            tokens,
            vec![
                Token::SlWhitespace("".to_string()),
                Token::Name("text".to_string()),
                Token::Class(".title".to_string()),
                Token::SlWhitespace(" ".to_string()),
                Token::LBrace("{".to_string()),
                Token::MlWhitespace("\n    ".to_string()),
                Token::SlComment("// text { \"hello world\" }".to_string()),
                Token::MlWhitespace("\n    ".to_string()),
                Token::String("\"hello world\"".to_string()),
                Token::MlWhitespace("\n".to_string()),
                Token::RBrace("}".to_string()),
            ]
        )
    }
}
